"""
Defines 3 sets of actions that are treated specially by the GitHub Actions
Reproducer.

`UNSUPPORTED_ACTIONS`:
For actions that (1) don't work with the current version of the Reproducer (e.g.
require secrets or tokens) and (2) are likely to cause jobs to fail if they're
just skipped.
`generate_pair_input.py` by default will exclude job pairs if their failed step
uses one of these actions.

`SKIPPED_ACTIONS`:
For actions that don't work with the current version of the Reproducer but can
usually be safely skipped.
The Reproducer will skip steps that use these actions.

`SPECIAL_ACTIONS`:
The reproducer will not download the action from GitHub and will replace it with
a custom action from the reproducer/resources directory.
"""

UNSUPPORTED_ACTIONS = {
    'actions/download-artifact',
}


SKIPPED_ACTIONS = UNSUPPORTED_ACTIONS | {
    'codecov/codecov-action',
    'actions/upload-artifact',
    'actions/download-artifact',
    'actions/cache',
    'gradle/wrapper-validation-action',
    'styfle/cancel-workflow-action',
    'github/codeql-action',
    'peaceiris/actions-gh-pages',
    's0/git-publish-subdir-action',
    'madrapps/jacoco-report',
    'mikepenz/action-junit-report',
}


SPECIAL_ACTIONS = {'actions/checkout'}



import time
from collections import deque

import cachecontrol
import requests

from bugswarm.common import log
from bugswarm.common import credentials

_BASE_URL = 'https://api.travis-ci.com'
# Number of seconds to sleep before retrying. Five seconds has been long enough to obey the Travis API rate limit.
_SLEEP_SECONDS = 5
_TOKENS = deque(credentials.TRAVIS_TOKENS)


class TravisWrapper(object):
    def __init__(self):
        self._session = cachecontrol.CacheControl(requests.Session())
        if credentials.TRAVIS_TOKENS:
            if not isinstance(credentials.TRAVIS_TOKENS, list):
                raise TypeError('TRAVIS_TOKENS must be a list.')
            if not all(isinstance(t, str) for t in credentials.TRAVIS_TOKENS):
                raise ValueError('All Travis CI Tokens must be given as strings.')

            # Start with the first token in TRAVIS_TOKENS list and pop() element from right and append to front
            # In the case where we are using 2 or more threads, each thread will grab the next token in the list
            self._session.headers['Authorization'] = 'token {}'.format(_TOKENS[0])
            _TOKENS.appendleft(_TOKENS.pop())

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._close()

    def _close(self):
        self._session.close()

    # Potentially raises requests.exceptions.Timeout or requests.exceptions.RequestException.
    def _get(self, address, **kwargs):
        sleep_seconds = _SLEEP_SECONDS
        attempts = 0
        while True:
            response = self._session.get(address, params=kwargs)
            code = response.status_code
            if code == 200:
                return response.json()
            elif code == 404:
                log.error('Get request for {} returned 404 Not Found.'.format(address))
                response.raise_for_status()
            elif code == 403:
                # Token did not successfully authorize, try next one in list
                # deque.pop() removes element from the right so we appendleft()
                self._session.headers['Authorization'] = 'token {}'.format(_TOKENS[0])
                _TOKENS.appendleft(_TOKENS.pop())
            elif code == 429:
                if attempts < 1 or not _TOKENS:
                    log.warning(
                        'The Travis API returned status code 429 Too Many Requests. '
                        'Retrying after sleeping for {} seconds.'.format(sleep_seconds))
                    time.sleep(sleep_seconds)
                    attempts += 1
                else:
                    # Use another token if # of attempts for GET Requests >= 1, will use next token in list
                    self._session.headers['Authorization'] = 'token {}'.format(_TOKENS[0])
                    _TOKENS.appendleft(_TOKENS.pop())
            else:
                log.error('Get request for {} returned {}.'.format(address, code))
                raise requests.exceptions.ConnectionError('{} download failed. Error code is {}.'.format(address, code))

    def _get_iterate(self, address, **kwargs):
        after_number = None
        build_number_exists = False
        if 'last_build_number' in kwargs:
            build_number = kwargs['last_build_number']
            build_number_exists = True
        result = self._get(address, **kwargs)
        latest_result_build_number = result[0]['number']
        if build_number_exists:
            if int(latest_result_build_number) == build_number:
                return
        while True:
            if after_number:
                if build_number_exists and int(after_number) < build_number:
                    return
                result = self._get(address, after_number=after_number)
            if not result:
                return
            yield from result
            after_number = result[-1]['number']
            if after_number == '1':
                return

    @staticmethod
    def _endpoint(path):
        return '{}/{}'.format(_BASE_URL, path)

    def search(self, term):
        return self._get_iterate(TravisWrapper._endpoint('search/repositories'), query=term)

    def get_builds_for_repo(self, repo, build_number=None):
        if build_number:
            return self._get_iterate(TravisWrapper._endpoint('repositories/{}/builds'.format(repo)),
                                     last_build_number=build_number)
        return self._get_iterate(TravisWrapper._endpoint('repositories/{}/builds'.format(repo)))

    def get_build_info(self, build_id):
        return self._get(TravisWrapper._endpoint('builds/{}'.format(build_id)))

    def get_job_info(self, job_id):
        return self._get(TravisWrapper._endpoint('jobs/{}'.format(job_id)))



import logging
import sys

from typing import Optional


def debug(*args):
    logging.getLogger().debug(_log_string(*args))


def info(*args):
    logging.getLogger().info(_log_string(*args))


def warning(*args):
    logging.getLogger().warning(_log_string(*args))


def error(*args):
    logging.getLogger().error(_log_string(*args))


def critical(*args):
    logging.getLogger().critical(_log_string(*args))


def _log_string(*args):
    str_args = [str(arg) for arg in args]
    return ' '.join(str_args)


def config_logging(log_level: int, log_file_path: Optional[str] = None, show_thread_name=False):
    logger = logging.getLogger()
    logger.handlers = []  # Remove previously added handlers before adding new ones.

    if show_thread_name:
        formatter = logging.Formatter('[%(levelname)8s] --- %(threadName)s %(message)s')
    else:
        formatter = logging.Formatter('[%(levelname)8s] --- %(message)s')

    if log_file_path:
        file_handler = logging.FileHandler(log_file_path, mode='w')
        file_handler.setFormatter(formatter)
        file_handler.setLevel(log_level)
        logger.addHandler(file_handler)

    stream_handler = logging.StreamHandler(stream=sys.stdout)
    stream_handler.setFormatter(formatter)
    stream_handler.setLevel(log_level)
    logger.addHandler(stream_handler)

    logger.setLevel(log_level)



import json
import sys
import os
import argparse
import git
import subprocess
import shutil
import re

import charset_normalizer

from bugswarm.common.rest_api.database_api import DatabaseAPI
from bugswarm.common import log
from bugswarm.common.credentials import DATABASE_PIPELINE_TOKEN


def _get_input_parser():
    parser = argparse.ArgumentParser(description='Calculate diff from BugSwarm aritfacts')
    required = parser.add_argument_group('required arguments')
    parser.add_argument('-t', '--image-tag',
                        default=None, help='image tag of the target BugSwarm artifact')
    parser.add_argument('-f', '--input-file',
                        default=None, help='Provide the input-file')
    required.add_argument('-j', '--json-path',
                          required=True, help='Please provide data json path')
    required.add_argument('-p', '--project-clone-path',
                          required=True, help='Please provide project clone path')
    args = parser.parse_args()
    image_tag = args.image_tag
    input_file = args.input_file
    project_clone_path = args.project_clone_path
    json_path = args.json_path

    if not os.path.exists(project_clone_path):
        os.makedirs(project_clone_path)

    if image_tag is not None and input_file is not None:
        print('Please provide image tag or input file which contains image tags')
        sys.exit(1)
    return image_tag, input_file, json_path, project_clone_path


# Cloning repo from github to a folder
def _git_clone_to_folder(project_clone_path, user, project):
    if not os.path.exists(project_clone_path):
        os.makedirs(project_clone_path)
    log.info('cloning repo... {}'.format(project))
    git.Repo.clone_from('https://github.com/{}'.format(user+'/'+project),
                        project_clone_path, odbt=git.GitDB)
    log.info('cloning done')


# Creating passed and failed folder of the repo
def _create_pass_fail_folder(project_clone_path, project):
    passed_path = os.path.join(project_clone_path, 'passed', project)

    if not os.path.exists(passed_path):
        os.makedirs(passed_path)

    src_path = os.path.join(project_clone_path, project)
    shutil.copytree(src_path, passed_path, dirs_exist_ok=True)


# revert to passed and failed sha using git to get passed and failed version
def _git_reset_and_fetch(folder_path, commit_sha):
    try:
        repo = git.Repo(folder_path)
        origin = repo.remote('origin')
        origin.fetch(commit_sha)

        repo.git.reset('--hard', commit_sha)
        log.info('git reset to ', commit_sha)
    except git.GitCommandError as e:
        log.info('An error has occurred ', e)
        raise Exception


def _get_diff_list(passed_commit, failed_commit, passed_path):
    cmd = 'cd {} && git diff {} {}'.format(
        passed_path,
        failed_commit,
        passed_commit
    )
    diff_list = []
    df = []

    result = subprocess.run(cmd, shell=True, capture_output=True, check=True)
    res_stdout = str(charset_normalizer.from_bytes(result.stdout).best())
    res_stderr = str(charset_normalizer.from_bytes(result.stderr).best())
    ok = result.returncode == 0

    if ok:
        result_lines = res_stdout.splitlines()
        for i in range(len(result_lines)):
            line = result_lines[i]
            if line.startswith('diff --git'):
                if i != 0:
                    diff_list.append(df)
                    df = []
            df.append(line)
            if i == len(result_lines) - 1:
                diff_list.append(df)
        return diff_list
    else:
        log.info('Could not fetch failed commit')
        log.error(res_stderr)
        return []


def _fill_up_patch(patch, old_file, new_file, content, added_code_size, deleted_code_size):
    patch['old_file'] = old_file
    patch['new_file'] = new_file
    patch['content'] = content
    patch['added_code_size'] = added_code_size
    patch['deleted_code_size'] = deleted_code_size
    return patch


def _create_patches(diff_list):
    diff_data = []
    plus_count = 0
    minus_count = 0
    for diff in diff_list:
        p = {}
        content = ''
        match = re.match(r'diff --git a/(.*) b/(.*)', diff[0])
        old_file = match.group(1)
        new_file = match.group(2)
        compare_line_index = 1

        if diff[compare_line_index].startswith('old mode'):
            compare_line_index = 3
            content += '{}\n{}\n'.format(diff[1], diff[2])

        # can happen if e.g. a file's mode was changed but its contents weren't
        if len(diff) <= compare_line_index:
            p = _fill_up_patch(p, old_file, new_file, content, 0, 0)

        # modified files
        elif diff[compare_line_index].startswith('index'):
            plus_per_file = 0
            minus_per_file = 0
            tripple_minus_or_binary_idx = compare_line_index + 1
            tripple_plus_idx = compare_line_index + 2
            content_start_idx = compare_line_index + 3
            if diff[tripple_minus_or_binary_idx].startswith('---') and diff[tripple_plus_idx].startswith('+++'):
                content = content + diff[tripple_minus_or_binary_idx] + '\n' + diff[tripple_plus_idx] + '\n'
                for i in range(content_start_idx, len(diff)):
                    content = content + diff[i] + '\n'
                    if diff[i].startswith('+'):
                        plus_count = plus_count + 1
                        plus_per_file = plus_per_file + 1
                    elif diff[i].startswith('-'):
                        minus_count = minus_count + 1
                        minus_per_file = minus_per_file + 1
                p = _fill_up_patch(p, old_file, new_file, content, plus_per_file, minus_per_file)
            # Binary Files - modified
            elif diff[tripple_minus_or_binary_idx].startswith('Binary'):
                content = 'Binary file {} modified'.format(old_file)
                p = _fill_up_patch(p, old_file, new_file, content, 0, 0)
            else:
                raise Exception('This diff is invalid due to {}'.format(diff[tripple_minus_or_binary_idx]))

        # added files
        elif diff[compare_line_index].startswith('new file'):
            plus_per_file = 0
            minus_per_file = 0
            p['old_file'] = '/dev/null'
            # added empty files
            if len(diff) == 3:
                content = 'Empty file.'
                p = _fill_up_patch(p, '/dev/null', new_file, content, 0, 0)
            else:
                # added files with contents
                tripple_minus_or_binary_idx = 3
                tripple_plus_idx = 4
                content_start_idx = 5
                if diff[tripple_minus_or_binary_idx].startswith('---') and diff[tripple_plus_idx].startswith('+++'):
                    content = content + diff[tripple_minus_or_binary_idx] + '\n' + diff[tripple_plus_idx] + '\n'
                    for i in range(content_start_idx, len(diff)):
                        content = content + diff[i] + '\n'
                        if diff[i].startswith('+'):
                            plus_count = plus_count + 1
                            plus_per_file = plus_per_file + 1
                    p = _fill_up_patch(p, '/dev/null', new_file, content, plus_per_file, 0)
                elif diff[tripple_minus_or_binary_idx].startswith('Binary'):
                    content = 'Binary file {} added'.format(new_file)
                    p = _fill_up_patch(p, '/dev/null', new_file, content, 0, 0)
                else:
                    raise Exception('This diff is invalid due to {}'.format(diff[tripple_minus_or_binary_idx]))

        # deleted files
        elif diff[compare_line_index].startswith('deleted'):
            plus_per_file = 0
            minus_per_file = 0
            if len(diff) == 3:
                content = 'Empty file.'
                p = _fill_up_patch(p, old_file, '/dev/null', content, 0, 0)
            else:
                tripple_minus_or_binary_idx = 3
                tripple_plus_idx = 4
                content_start_idx = 5
                if diff[tripple_minus_or_binary_idx].startswith('---') and diff[tripple_plus_idx].startswith('+++'):
                    # deleted files with contents
                    content = content + diff[tripple_minus_or_binary_idx] + '\n' + diff[tripple_plus_idx] + '\n'
                    for i in range(content_start_idx, len(diff)):
                        content = content + diff[i] + '\n'
                        if diff[i].startswith('-'):
                            minus_count = minus_count + 1
                            minus_per_file = minus_per_file + 1
                    p = _fill_up_patch(p, old_file, '/dev/null', content, 0, minus_per_file)
                elif diff[tripple_minus_or_binary_idx].startswith('Binary'):
                    # Binary files /dev/null and b/po.zip differ
                    content = 'Binary file {} deleted'.format(old_file)
                    p = _fill_up_patch(p, old_file, '/dev/null', content, 0, 0)
                else:
                    raise Exception('This diff is invalid due to {}'.format(diff[tripple_minus_or_binary_idx]))

        # renaming
        elif diff[compare_line_index].startswith('similarity'):
            # similarity index 100%
            plus_per_file = 0
            minus_per_file = 0
            match1 = re.match(r'similarity index (.*)%', diff[compare_line_index])
            match_percentage = int(match1.group(1))
            o = old_file
            n = new_file
            tripple_minus_or_binary_idx = compare_line_index + 4
            tripple_plus_idx = compare_line_index + 5
            content_start_idx = compare_line_index + 6
            if match_percentage == 100:
                content = 'File renamed from {} to {}'.format(o, n)
                p = _fill_up_patch(p, old_file, new_file, content, 0, 0)
            else:
                if diff[tripple_minus_or_binary_idx].startswith('Binary'):
                    content = 'File renamed from {} to {}'.format(o, n)
                    p = _fill_up_patch(p, old_file, new_file, content, 0, 0)
                else:
                    content = content + diff[tripple_minus_or_binary_idx] + '\n' + diff[tripple_plus_idx] + '\n'
                    for i in range(content_start_idx, len(diff)):
                        content = content + diff[i] + '\n'
                        if diff[i].startswith('+'):
                            plus_count = plus_count + 1
                            plus_per_file = plus_per_file + 1
                        elif diff[i].startswith('-'):
                            minus_count = minus_count + 1
                            minus_per_file = minus_per_file + 1
                    p = _fill_up_patch(p, old_file, new_file, content, plus_per_file, minus_per_file)

        else:
            raise Exception('This diff is invalid due to {}'.format(diff[compare_line_index]))

        diff_data.append(p)

    return diff_data, plus_count, minus_count


# generate diff entry for the schema
def _generate_diff_input(image_tag, input_file, json_file_path, project_clone_path):
    os.environ['GIT_TERMINAL_PROMPT'] = '0'  # So commands that normally ask for credentials auto-fail.
    image_tag_list = []

    if image_tag is None:
        with open(input_file, 'r') as f:
            lines = f.readlines()

        for line in lines:
            image_tag_list.append(line.split('\n')[0])

    elif input_file is None:
        image_tag_list.append(image_tag)

    bugswarmapi = DatabaseAPI(token=DATABASE_PIPELINE_TOKEN)
    diff_data = []
    errored_tags = []
    for image_tag in image_tag_list:
        response = bugswarmapi.find_artifact(image_tag)
        if not response.ok:
            log.error('Artifact not found: {}'.format(image_tag))
            continue

        diff_entry = {}
        artifact = response.json()
        diff_entry['image_tag'] = artifact['image_tag']
        diff_entry['failed_sha'] = artifact['failed_job']['trigger_sha']
        diff_entry['passed_sha'] = artifact['passed_job']['trigger_sha']
        user = artifact['repo'].split('/')[0]
        project = artifact['repo'].split('/')[1]
        passed_path = os.path.join(project_clone_path, 'passed', project)

        try:
            _git_clone_to_folder(project_clone_path, user, project)
            _create_pass_fail_folder(project_clone_path, project)

            _git_reset_and_fetch(passed_path, diff_entry['passed_sha'])
            _git_reset_and_fetch(passed_path, diff_entry['failed_sha'])
            diffs = _get_diff_list(diff_entry['passed_sha'], diff_entry['failed_sha'], passed_path)

            diff_, plus, minus = _create_patches(diffs)
            diff_size = plus + minus
            diff_entry['patches'] = diff_
            diff_entry['diff_size'] = diff_size
            diff_entry['total_added_code'] = plus
            diff_entry['total_deleted_code'] = minus
            diff_data.append(diff_entry)

        except Exception as e:
            log.logging.exception('Skipping {} due to an error {}:'.format(image_tag, e))
            errored_tags.append(image_tag)

        shutil.rmtree(passed_path, ignore_errors=True)
        project_path_to_del = os.path.join(project_clone_path, project)
        shutil.rmtree(project_path_to_del, ignore_errors=True)

    with open(json_file_path, 'w') as f:
        json.dump(diff_data, f)
    with open('error_image.txt', 'w') as f:
        f.write('\n'.join(errored_tags) + '\n')


def gather_diff_info(failed_sha, passed_sha, repo, url):
    user = repo.split('/')[0]
    project = repo.split('/')[1]
    repo_clone_path = os.path.join(os.path.dirname(os.getcwd()), 'github-pair-finder', 'intermediates', 'repos',
                                   repo.replace('/', '-'))
    log.info(repo_clone_path)
    if not os.path.exists(repo_clone_path):
        _git_clone_to_folder(repo_clone_path, user, project)

    info = {}
    info['url'] = url
    try:
        _git_reset_and_fetch(repo_clone_path, passed_sha)
        _git_reset_and_fetch(repo_clone_path, failed_sha)
        diffs = _get_diff_list(passed_sha, failed_sha, repo_clone_path)
        diff_, plus, minus = _create_patches(diffs)
        changed_files = []
        for d in diff_:
            changed_files.append(d['new_file'])
        diff_size = plus + minus
        info['metrics'] = {
            'num_of_changed_files': len(changed_files),
            'changes': diff_size,
            'additions': plus,
            'deletions': minus
        }
        info['changed_paths'] = changed_files
        info['error_found'] = 'NONE'
    except Exception as e:
        info['num_of_changed_files'] = -1
        info['changed_paths'] = ['ERROR, CANNOT FULFILL REQUEST']
        info['error_found'] = 'ERROR, {}'.format(e)
        info['metrics'] = {
            'num_of_changed_files': 0,
            'changes': 0,
            'additions': 0,
            'deletions': 0
        }
        log.logging.exception('Could not gather info of {} {}..{} due to an error:'.format(repo, failed_sha,
                                                                                           passed_sha))
    return info



"""
Defines decorators that enable class properties.

These decorators are adapted from answers to the Stack Overflow question at https://stackoverflow.com/q/3203286.
"""


class classproperty:
    """
    Same as property(), but passes obj.__class__ instead of obj to fget/fset/fdel.
    Original code for property emulation at https://docs.python.org/3.5/howto/descriptor.html#properties.
    """

    def __init__(self, fget=None, fset=None, fdel=None, doc=None):
        self.fget = fget
        self.fset = fset
        self.fdel = fdel
        if doc is None and fget is not None:
            doc = fget.__doc__
        self.__doc__ = doc

    def __get__(self, obj, objtype=None):
        if obj is None:
            return self
        if self.fget is None:
            raise AttributeError('unreadable attribute')
        return self.fget(obj.__class__)

    def __set__(self, obj, value):
        if self.fset is None:
            raise AttributeError("can't set attribute")
        self.fset(obj.__class__, value)

    def __delete__(self, obj):
        if self.fdel is None:
            raise AttributeError("can't delete attribute")
        self.fdel(obj.__class__)

    def getter(self, fget):
        return type(self)(fget, self.fset, self.fdel, self.__doc__)

    def setter(self, fset):
        return type(self)(self.fget, fset, self.fdel, self.__doc__)

    def deleter(self, fdel):
        return type(self)(self.fget, self.fset, fdel, self.__doc__)


def classproperty_support(cls):
    """
    Class decorator to add metaclass to our class. Metaclass uses to add descriptors to class attributes.
    """
    # Use type(cls) to use metaclass of given class
    class Meta(type(cls)):
        pass

    for name, obj in vars(cls).items():
        if isinstance(obj, classproperty):
            setattr(Meta, name, property(obj.fget, obj.fset, obj.fdel))

    class Wrapper(cls, metaclass=Meta):
        pass

    return Wrapper



import re
import subprocess

from distutils.version import StrictVersion
from typing import List

import requests

from bugswarm.common import log
from .shell_wrapper import ShellWrapper


def check_package_outdated(package: str):
    """
    Checks if the installed version of a package is older than the latest non-prerelease version available on PyPI.
    If so, prints a message the asks the user to consider upgrading.

    The package must be available on PyPI and must have always used a version numbering scheme that can be parsed by
    distutils.version.StrictVersion.

    This function is meant to be used for packages in the 'bugswarm' namespace, which meet the above requirements, and
    therefore is not guaranteed to work for packages outside that namespace.

    :param package: The name of the package to check.
    """
    if not isinstance(package, str):
        raise TypeError

    try:
        installed = _get_installed_version(package)
        latest = _get_latest_version(package)
        if latest > installed:
            # A newer, non-prerelease version is available.
            log.info('You are using {} version {}, but version {} is available.'.format(package, installed, latest))
            log.info("You should consider upgrading via the 'pip3 install --upgrade {}' command.".format(package))
    except Exception as e:
        log.error('Encountered an error while checking if {} can be updated: {}'.format(package, e))


def _get_installed_version(package: str) -> StrictVersion:
    if not isinstance(package, str):
        raise TypeError

    stdout, _, returncode = ShellWrapper.run_commands('pip show {}'.format(package), stdout=subprocess.PIPE, shell=True)
    version_string = re.search(r'Version: (.*)', stdout, re.IGNORECASE).group(1)
    return StrictVersion(version_string)


def _get_latest_version(package: str) -> StrictVersion:
    if not isinstance(package, str):
        raise TypeError

    return sorted(_list_non_prerelease_versions(package), reverse=True)[0]


def _list_non_prerelease_versions(package: str) -> List[StrictVersion]:
    """
    Returns a list of StrictVersion objects representing the available, non-prerelease versions of a package, if the
    package is available on PyPI. Otherwise, returns an empty list.
    """
    url = _pypi_api_url(package)
    resp = requests.get(url)
    all_versions = list(map(StrictVersion, resp.json()['releases'].keys()))
    non_prerelease_versions = [v for v in all_versions if not v.prerelease]
    return non_prerelease_versions


def _pypi_api_url(package: str):
    return 'https://pypi.org/pypi/{}/json'.format(package)



"""
Convenience methods and constants to facilitate a processing workflow involving BugSwarm artifact images.
"""
import os
import shutil

from ..shell_wrapper import ShellWrapper
from ..rest_api.database_api import DatabaseAPI

REPOS_DIR = '/home/travis/build'
_SANDBOX = 'bugswarm-sandbox'
HOST_SANDBOX = os.path.join(os.path.expanduser('~'), _SANDBOX)
CONTAINER_SANDBOX = os.path.join(os.sep, _SANDBOX)


def copy_to_host_sandbox(filepath: str):
    """
    Copy a file into the host-side sandbox.
    :param filepath: Path to the file to be copied.
    """
    if not filepath:
        raise ValueError
    if not os.path.isfile(filepath):
        raise FileNotFoundError
    shutil.copy(filepath, HOST_SANDBOX)


def get_repo(image_tag: str, token: str):
    """
    Get the repository slug for the artifact represented by `image_tag`.
    """
    if not image_tag:
        raise ValueError
    bugswarmapi = DatabaseAPI(token=token)
    resp = bugswarmapi.find_artifact(image_tag)
    resp.raise_for_status()
    return resp.json()['repo']


def get_failed_repo_dir(image_tag: str, token: str):
    """
    Get the path to the failed repository in the container.
    """
    if not image_tag:
        raise ValueError
    return os.path.join(REPOS_DIR, 'failed', *get_repo(image_tag, token).split('/'))


def get_passed_repo_dir(image_tag: str, token: str):
    """
    Get the path to the passed repository in the container.
    """
    if not image_tag:
        raise ValueError
    return os.path.join(REPOS_DIR, 'passed', *get_repo(image_tag, token).split('/'))


def run_artifact(image_tag: str, command: str):
    """
    Assumes that the caller wants to use the sandbox and stdin piping features of the BugSwarm client since this
    function will likely be called in the context of an artifact processing workflow.

    :param image_tag: The image tag representing the artifact image to run.
    :param command: A string containing command(s) to execute in the artifact container. Will be piped to the container
                    process' standard input stream.
    :return: A 2-tuple of
             - the combined output of stdout and stderr
             - the return code of the subprocess that ran the artifact container.
    """
    if not image_tag:
        raise ValueError
    if not command:
        raise ValueError
    combined, _, returncode = ShellWrapper.run_commands(
        'bugswarm run --image-tag {} --use-sandbox --pipe-stdin'.format(image_tag),
        input=command,
        universal_newlines=True,
        shell=True)
    return combined, returncode



"""
String constants for each reason that Pair Filter filters a pair. Also used in the evaluation script.

Think carefully before changing these constants! Any changes will create inconsistency in the 'filtered_reason'
attribute in the metadata database. In turn, changes will break the evaluation script. In short, if these constants
change, then to remove inconsistencies between previously created artifacts and newly created artifacts, either every
artifact will need to be recreated or the metadata database will need to be updated so that previously created artifacts
use the new constants in their metadata.
"""

NO_HEAD_SHA = 'no head sha'
NO_ORIGINAL_LOG = 'do not have original log'
ERROR_READING_ORIGINAL_LOG = 'error when reading original log'
NO_IMAGE_PROVISION_TIMESTAMP = 'original log does not have provisioned datetime string'
INACCESSIBLE_IMAGE = 'do not have image'
NOT_RESETTABLE = 'not resettable'  # Deprecated.
NOT_ACQUIRABLE = 'not acquirable'  # Deprecated.
NOT_AVAILABLE = 'not available'
SAME_COMMIT_PAIR = 'failed build has same sha with passed build'
UNAVAILABLE_RUNNER = 'do not have github runner'
UNRESETTABLE_WITH_SUBMODULES = 'repo is non-resettable and has submodules'
UNREDACTED_TOKEN = 'job config includes unredacted/cleartext tokens'
UNSUPPORTED_WORKFLOW = 'job workflow file uses unsupported syntax'
FIRST_STEP_NOT_CHECKOUT = 'first step is not actions/checkout or uses unsupported parameters'
JOBS_FROM_DIFFERENT_PRS = 'failed and passed jobs are from different PRs'
ORIGINAL_LOG_TOO_LARGE = 'original log is too large'
JOB_FAILED_DURING_CHECKOUT = 'failed job failed during the checkout action'



import os
import time
import urllib.request

from concurrent.futures import as_completed
from concurrent.futures import ThreadPoolExecutor
from typing import List
from typing import Union
from urllib.error import URLError
from urllib.error import HTTPError

from bugswarm.common import log
from bugswarm.common.credentials import GITHUB_TOKENS
from bugswarm.common.github_wrapper import GitHubWrapper

_DEFAULT_RETRIES = 3


def download_log(job_id: Union[str, int],
                 destination: str,
                 overwrite: bool = True,
                 retries: int = _DEFAULT_RETRIES,
                 repo: str = '') -> bool:
    """
    Downloads a Travis/GitHub job log and stores it at destination.

    :param job_id: Travis/GitHub job ID, as a string or integer, identifying a job whose log to download.
    :param destination: Path where the log should be stored.
    :param overwrite: Whether the file at `destination`, if it already exists, should be replaced.
    :param retries: The number of times to retry the log download if it fails. Defaults to 3, in which case the network
                    will be accessed up to 4 (3 + 1) times.
    :param overwrite: Whether to overwrite a file at `destination` if one already exists. If `overwrite` is False and a
                      file already exists at `destination`, FileExistsError is raised. Defaults to True.
    :param repo: The repository of the job. For GitHub job only, set it to empty string to download Travis log.
    :raises ValueError:
    :raises FileExistsError: When a file at `destination` already exists and `overwrite` is False.
    :return: True if the download succeeded.
    """
    if not job_id:
        raise ValueError
    if not destination:
        raise ValueError
    if os.path.isfile(destination) and not overwrite:
        log.error('The log for job', job_id, 'already exists locally.')
        raise FileExistsError

    job_id = str(job_id)

    if repo == '':
        # Travis log
        travis_log_link = 'https://api.travis-ci.org/jobs/{}/log.txt'.format(job_id)
        content = _get_log_from_url(travis_log_link, retries)

        if not content:
            travis_log_link = 'https://api.travis-ci.com/v3/job/{}/log.txt'.format(job_id)
            content = _get_log_from_url(travis_log_link, retries)
            # If this endpoint fails, the log is not on either endpoint and does not exist
            if not content:
                return False
    else:
        # GitHub log
        github_log_link = 'https://api.github.com/repos/{}/actions/jobs/{}/logs'.format(repo, job_id)
        content = _get_github_log_from_url(github_log_link, retries)
        if not content:
            return False

    with open(destination, 'wb') as f:
        f.write(content)
    return True


def download_logs(job_ids: List[Union[str, int]],
                  destinations: List[str],
                  overwrite: bool = True,
                  num_workers: int = 5,
                  retries: int = _DEFAULT_RETRIES,
                  repos=None) -> bool:
    """
    Downloads one or more Travis/GitHub job logs in parallel and stores them at the given destinations.
    This function calls `download_log` and raises the first exception it catches from that function, if any.

    If you only need to download a single Travis job log, use the `download_log` function.

    :param job_ids: A list of Travis/GitHub job IDs, as strings or integers, identifying jobs whose logs to download.
    :param destinations: A list of paths where the logs should be stored. The path at index `i` corresponds to the log
                         downloaded for the job ID at index `i` in `job_ids`. Thus, `job_ids` and `destinations` must be
                         the same length.
    :param overwrite: Same as the argument for `download_log`.
    :param num_workers: Number of workers to download logs. Defaults to the maximum of 5.
    :param retries: Same as the argument for `download_log`.
    :param repos: A list of jobs' repository. For GitHub jobs only, use empty list to download Travis logs.
    :raises ValueError:
    :raises FileExistsError: When a file already exists at the given destination and `overwrite` is False.
    :return: True if all downloads succeeded.
    """
    if not job_ids:
        raise ValueError
    if not destinations:
        raise ValueError
    if not len(job_ids) == len(destinations):
        log.error('The job_ids and destinations arguments must be of equal length.')
        raise ValueError
    if repos is not None and len(repos) != len(job_ids):
        log.error('The job_ids and repositories arguments must be of equal length.')
        raise ValueError

    num_workers = min(num_workers, len(job_ids))
    with ThreadPoolExecutor(max_workers=num_workers) as executor:
        if repos is None:
            future_to_job_id = {executor.submit(download_log, job_id, dst, overwrite, retries): job_id
                                for job_id, dst in zip(job_ids, destinations)}
        else:
            future_to_job_id = {executor.submit(download_log, job_id, dst, overwrite, retries, repo): job_id
                                for job_id, dst, repo in zip(job_ids, destinations, repos)}

    succeeded = 0
    for future in as_completed(future_to_job_id):
        try:
            # The result will be True if the download succeeded. Otherwise, future.result() will raise an exception or
            # return False.
            ok = future.result()
        except Exception:
            raise
        else:
            if ok:
                succeeded += 1

    return succeeded == len(job_ids)


def _get_log_from_url(log_url: str, max_retries: int, retry_count: int = 0):
    sleep_duration = 3  # Seconds.
    try:
        req = urllib.request.Request(log_url)
        with urllib.request.urlopen(req) as url:
            result = url.read()
            log.info('Downloaded log from {}.'.format(log_url))
            return result
    except HTTPError as e:
        if e.code in [410, 500]:
            log.error('The log ({}) for this job have expired.'.format(log_url))
        else:
            log.error('Could not download log from {}, unexpected status code: {}'.format(log_url, e.code))
        return None
    except URLError as e:
        log.error('Could not download log from {}: {}.'.format(log_url, e.reason))
        return None
    except ConnectionResetError:
        if retry_count == max_retries:
            log.warning('Could not download log from', log_url, 'after retrying', max_retries, 'times.')
            return None
        log.warning('The server reset the connection. Retrying after', sleep_duration, 'seconds.')
        time.sleep(sleep_duration)
        _get_log_from_url(log_url, max_retries, retry_count + 1)


def _get_github_log_from_url(log_url: str, max_retries: int):
    github_wrapper = GitHubWrapper(GITHUB_TOKENS)

    response, log_text = github_wrapper.get(log_url, response_format='bytes', max_retry=max_retries)
    if not response.ok:
        if response.status_code in [410, 500]:
            log.error('The log ({}) for this job has expired.'.format(log_url))
        else:
            log.error('Could not download log from {}: status code {}'.format(log_url, response.status_code))
        return None
    return log_text
